#!/usr/bin/env python3
#
# Copyright (c) 2021, 2024, Oracle and/or its affiliates.
#

import socket
import sys
import os.path


class NoFreePort(Exception):
    pass

def PrintUsage(appname):
    print("Usage:")
    print("    " + appname + " <lock_prefix> <start> <end>")
    print("    " + appname + " <lock_prefix> -r|--remove <port>")

def GetLockName(prefix, port):
    """Create a lock filename for the given prefix and port
    param: str prefix: the prefix to use in lock filenames.
    param: int port: the port to create the lock file for.
    """
    return "/tmp/." + str(prefix) + str(port) + "-lock"

def CreateLock(prefix, port):
    """Create a lock file for the given prefix and port
    param: str prefix: the prefix to use in lock filenames.
    param: int port: the port to create the lock file for.
    """
    lock_file = GetLockName(prefix, port)
    f = open(lock_file, "w+")
    f.write(str(os.getpid()))
    f.close()

def DeleteLock(prefix, port):
    """Delete the lock file for the given prefix and port
    param: str prefix: the prefix to use in lock filenames.
    param: int port: the port to dlete the lock file for.
    """
    lock_file = GetLockName(prefix, port)
    try:
        os.remove(lock_file)
    except:
        pass

def FindPort(prefix, start, end):
    """Find the first bindable port in the range [start,end)
    param: str prefix: the prefix to use in lock filenames.
    param: int start:  the first port number we'll try to bind to.
    param: int end:    the final port number we'll try to bind to.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    step = 1 if start < end else -1

    for port in range(start, end+1, step):
        lock_file = GetLockName(prefix, port)
        if os.path.exists(lock_file):
            # Skip this port as there is a lock file for it
            continue;
        try:
            # '' is a special case and represents INADDR_ANY
            sock.bind(('', port))
        except socket.error:
            continue
        else:
            # Create a lock file for this port
            CreateLock(prefix, port);
            return str(port)

    raise NoFreePort

if __name__ == "__main__":
    if len(sys.argv) != 4:
        PrintUsage(sys.argv[0])
        sys.exit(1)
    
    prefix = sys.argv[1]
    if sys.argv[2] == "-r" or sys.argv[2] == "--remove":
        # Delete the lock file for this port
        DeleteLock(prefix, sys.argv[3])
        sys.exit(0)

    # Otherwise find a port
    start = int(sys.argv[2])
    end   = int(sys.argv[3])

    try:
        port = FindPort(prefix, start, end)
    except NoFreePort as e:
        sys.exit(1)
    
    print(port)

    sys.exit(0)


