#!/bin/bash
#
# Copyright (c) 2021, 2024, Oracle and/or its affiliates.
#

set -o nounset -o errexit -o pipefail
set -m

trap '' SIGHUP

SPRITE_X11_DIR_DEFAULT=/opt/oracle/sprite-x11

SPRITE_X11_DIR=${SPRITE_X11_HOME:-${SPRITE_X11_DIR_DEFAULT}}
SPRITE_X11_BIN="${SPRITE_X11_DIR}/bin"

# Pull in common config
source "${SPRITE_X11_BIN}/common.sh"

# Pull in logging levels
source "${SPRITE_X11_BIN}/logging.sh"

# Pull in statuses
source "${SPRITE_X11_BIN}/status.sh"

XVFB_MAXWIDTH=8192
XVFB_MAXHEIGHT=4096
XVFB_WIDTH=1920
XVFB_HEIGHT=1080
XVFB_ARGS="-screen 0 ${XVFB_MAXWIDTH}x${XVFB_MAXHEIGHT}x24"
XSERVER_PORT_FILE="${HOME}/.sprite-x11-xport"

XORG_BIN="/usr/bin/Xorg"
XORG_ARGS="-r -nolisten tcp -noreset +extension DAMAGE +extension RANDR +extension RENDER +extension DPMS -config sprite-xorg.conf"


XAUTH_FILE="${HOME}/.Xauthority"
X_PORT="1"

AUDIO_SAMPLE_RATE=0
AUDIO_PORT_FILE="${HOME}/.sprite-x11-audioport"
AUDIO_PORT="9775"

XSERVER="xorg"
XSERVER_OUTPUT="DUMMY0"

display_usage()
{
    echo "Usage:"
    echo "  x11AppService.sh [OPTION…] -- <command> [args]"
    echo ""
    echo "Options:"
    echo "  -h|--help               This help message."
    echo "  -W|--width <width>      Width in pixels."
    echo "  -H|--height <height>    Height in pixels."
    echo "  -f|--xauth-file <FILE>  Xauth file."
    echo "  -a|--audio-rate <rate>  Audio sample rate (samples/sec)."
    echo "  -x|--xserver <server>   XServer to use, \"xorg\" or \"xvfb\"."
    echo ""
    echo "Arguments:"
    echo "  <command>       Application command to run. Required."
    echo "  <args>          Arguments to application command. Optional."
}

exit_failure()
{
    local errorCode=${1}
    local errorString="${STATUSES[${errorCode}]}"
    echo "status:${errorString}"
    exit ${errorCode}
}

terminate()
{
    echo "x11AppService terminate" >> ${LOGFILE} 2>&1
    date >> ${LOGFILE} 2>&1

    trap '' CHLD
    set +e
    remove_xauth_cookie
    # Kill any process that might be left
    kill "${XSERVER_PID}" &>/dev/null
    kill "${APPLICATION_PID}" &>/dev/null
    kill "${AUDIOSERVER_PID}" &> /dev/null
    # Remove any lock files
    ${SPRITE_X11_BIN}/findport.py audioserver -r ${AUDIO_PORT}
    rm -rf "${XSERVER_PORT_FILE}"
    rm -rf "${AUDIO_PORT_FILE}"
    set -e
}

start_xserver()
{
    local serverType="xorg"
    if [ $# -gt 0 ]
    then
        serverType="$1"
    fi
    
    if [ "${SPRITE_X11_SINGLE_INSTANCE}" == "1" ]
    then
        # Fix the Xport to use
        XPORT_ARGS=":${X_PORT}"
        echo "${X_PORT}" > ${XSERVER_PORT_FILE}
    else
        # Let Xvfb find a free port
        XPORT_ARGS="-displayfd ${XSERVER_DISPLAYFD}"
    fi

    echo "start_xserver ${serverType} ${XPORT_ARGS}" >>${LOGFILE} 2>&1

    if [ "${serverType}" == "xvfb" ]
    then
        "Xvfb" \
            ${XVFB_ARGS} \
            -nolisten tcp \
            -noreset \
            -r \
            -auth ${XAUTH_FILE} \
            ${XPORT_ARGS} \
            >>${LOGFILE} 2>&1 &
    else
        "${XORG_BIN}" \
            ${XORG_ARGS} \
            -auth ${XAUTH_FILE} \
            ${XPORT_ARGS} \
            >>${LOGFILE} 2>&1 &
    fi

    XSERVER_PID=$!
    echo "XServer pid ${XSERVER_PID}" >>${LOGFILE} 2>&1

    # Wait until the Xvfb server is running
    # for now just give it a bit more time
    sleep 2

    if [ "${SPRITE_X11_SINGLE_INSTANCE}" != "1" ]
    then
        # Retrieve the X port from the displayfd file
        X_PORT=$(cat <"${XSERVER_PORT_FILE}")
    fi
    echo "X port ${X_PORT}" >>${LOGFILE} 2>&1

    # Set the environment
    DISPLAY=":${X_PORT}"
    export DISPLAY
    XAUTHORITY=${XAUTH_FILE}
    export XAUTHORITY

    # Set the xauth cookie
    set_xauth_cookie
}

# Start the audio server
# Failure is not fatal
start_audio_server()
{
    AUDIOSERVER_PID=""

    if [ "${AUDIO_SAMPLE_RATE}" == "0" ]
    then
        return
    fi
 
    set +e
    if [ "${SPRITE_X11_SINGLE_INSTANCE}" != "1" ]
    then
        AUDIO_PORT="$(${SPRITE_X11_BIN}/findport.py audioserver 9775 9676)"
        if [ $? -ne 0 ]
        then
            set -e
            return
        fi
    fi

    echo "${AUDIO_PORT}" > "${AUDIO_PORT_FILE}"
    echo "audio_port ${AUDIO_PORT}" >>${LOGFILE} 2>&1

    AUDIOSERVER_PID=$(${SPRITE_X11_BIN}/audioService.sh -p ${AUDIO_PORT} -s ${AUDIO_SAMPLE_RATE})
    if [ $? -ne 0 ]
    then
        AUDIOSERVER_PID=""
        set -e
        return
    fi
    set -e
    echo "audioserver pid ${AUDIOSERVER_PID}" >>${LOGFILE} 2>&1
}

# Resize the screen to the preferred size
resize_Xserver()
{
    local MODENAME="Sprite_${1}x${2}"
    set +e
    xrandr --newmode ${MODENAME} 0.0 ${1} 0 0 0 ${2} 0 0 0 &> /dev/null
    xrandr --addmode ${XSERVER_OUTPUT} ${MODENAME} &> /dev/null
    xrandr --output ${XSERVER_OUTPUT} --mode ${MODENAME} &> /dev/null
    set -e
}

# Generate xauth cookie 
set_xauth_cookie()
{
    xauth add ${DISPLAY} "." $(mcookie) &> /dev/null
}

# Remove xauth cookie 
remove_xauth_cookie()
{
    xauth remove ${DISPLAY} &> /dev/null
}

start_app_layer()
{
    local serverType="$1"
    start_xserver "${serverType}"
    start_audio_server

    # Echo the ports out for use by the access layer
    echo "${X_PORT}" "${AUDIO_PORT}"

    # Resize the display
    resize_Xserver "${XVFB_WIDTH}" "${XVFB_HEIGHT}"

    # Run the application or desktop.
    "${APPLICATION}" ${APPLICATION_ARGS} &> /dev/null &
    APPLICATION_PID=$!
    echo "application pid ${APPLICATION_PID}" >>${LOGFILE} 2>&1

    wait ${APPLICATION_PID}

    terminate
}



echo "Starting AppService" >> ${LOGFILE}

set +e
options=$(getopt -q -o hW:H:f:a:x: -l help,width:,height:,xauth-file:,audio-rate:,xserver: -- "$@")
if [ $? -ne 0 ]
then
    exit_failure ${STATUS_E_UNKNOWNARG}
fi
set -e

eval set -- "$options"
while true; do
    case "$1" in
    -h|--help)
        display_usage
        exit 0
        ;;
    -W|--width)
        XVFB_WIDTH="$2"
        shift 
        ;;
    -H|--height)
        XVFB_HEIGHT="$2"
        shift 
        ;;
    -f|--xauth-file)
        XAUTH_FILE="$2"
        shift 
        ;;
    -a|--audio-rate)
        AUDIO_SAMPLE_RATE="$2"
        shift
        ;;
    -x|--xserver)
        XSERVER="$2"
        if [ "${XSERVER}" == "xorg" ]
        then 
            XSERVER_OUTPUT="DUMMY0"
        elif [ "${XSERVER}" == "xvfb" ]
        then
            XSERVER_OUTPUT="screen"
        else
            echo "XServer must be either \"xorg\" or \"xvfb\"."
            exit_failure ${STATUS_E_UNKNOWNARG}
        fi
        shift
        ;;
    --)
        shift
        break
        ;;
    esac
    shift
done

# Check there's at least an application command.
if [ $# -eq 0 ]
then
    exit_failure ${STATUS_E_NOAPPCOMMAND}
fi

# First positional parameter is the application command.
APPLICATION="$1"
# The remainder are arguments to that command.
APPLICATION_ARGS="${@:2}"
echo "application: ${APPLICATION=} args: ${APPLICATION_ARGS}" >> ${LOGFILE} 2>&1

if [ ! -x "${APPLICATION}" ]
then
    exit_failure ${STATUS_E_APPNOTFOUND}
fi

# Check for a file for Xvfb to use for displayfd
if [ -f "${XSERVER_PORT_FILE}" ]
then
    exec {XSERVER_DISPLAYFD}>> "${XSERVER_PORT_FILE}"
else
    exec {XSERVER_DISPLAYFD}> "${XSERVER_PORT_FILE}"
fi

# Lock our file
if (flock -xn "${XSERVER_DISPLAYFD}")
then
    echo "Create new app layer" >>${LOGFILE} 2>&1
    # No existing app layer session
    # Create one
    read X_PORT AUDIO_PORT < <( start_app_layer "${XSERVER}" & )
else
    echo "Connect existing app layer" >>${LOGFILE} 2>&1
    # Existing app layer session
    # Connect to the existing app layer session
    if [ "${SPRITE_X11_SINGLE_INSTANCE}" != "1" ]
    then
        X_PORT=$(cat <"${XSERVER_PORT_FILE}")
        if [ -f "${AUDIO_PORT_FILE}" -a "${AUDIO_SAMPLE_RATE}" -ne 0 ]
        then
            AUDIO_PORT=$(cat <"${AUDIO_PORT_FILE}")
        fi
    fi
    echo "X port ${X_PORT}, audio port ${AUDIO_PORT}" >>${LOGFILE} 2>&1
fi

echo "${X_PORT}" "${AUDIO_PORT}"

exit ${STATUS_S_OK}
